
// board data
function BoardData() {
    // ' ': empty
    // '*': filled
    // '-': marked
    this.width  = 0;
    this.height = 0;
    this.answer = "";
    this.dots   = "";
    this.title = null;
    this.author = null;

    this.horizontalNumbers = null;
    this.verticalNumbers   = null;
}

BoardData.prototype.isValid = function() {
    if ( this.width <= 0 || this.height <= 0 ) {
        return false;
    }
    if ( this.answer.length >= this.width * this.height ) {
        return true;
    }
    return false;
};
BoardData.prototype.completes = function() {
    for ( var i = 0; i < this.width; i++ ) {
        if ( this.columnMatches( i ) == false ) {
            return false;
        }
    }
    for ( var i = 0; i < this.height; i++ ) {
        if ( this.rowMatches( i ) == false ) {
            return false;
        }
    }
    return true;
};

BoardData.prototype.columnMatches = function( column ) {
    var numbers = this.verticalNumbers[ column ];
    var count   = 0;
    var num     = 0;
    var flag_correct = true;
    var pos = column;

    for ( var i = 0; i < this.height; i++, pos+= this.width ) {
        if ( this.dots[ pos ] != '*' ) {
            //if ( this.data.isChecked( col, i ) == false ) {
            if ( count > 0 ) {
                if ( num >= numbers.length || numbers[ num ] != count ) {
                    flag_correct = false;
                    break;
                }
                num ++;
                count = 0;
            }
        } else {
            count ++;
        }
    }

    if ( flag_correct ) {
        if ( count > 0 && ( num != numbers.length - 1 ) ) {
            flag_correct = false;
        } else if ( count > 0 && count != numbers[ num ] ) {
            flag_correct = false;
        } else if ( count == 0 && num < numbers.length ) {
            flag_correct = false;
        }
    }

    return flag_correct;
};

BoardData.prototype.rowMatches = function( row ) {
    var num = 0;
    var count = 0;
    var flag_correct = true;
    var pos = row * this.width;
    var numbers = this.horizontalNumbers[ row ];

    for ( var i = 0; i < this.width; i++, pos++ ) {
        if ( this.dots[ pos ] != '*' ) {
            if ( count > 0 ) {
                if ( num >= numbers.length || numbers[ num ] != count ) {
                    flag_correct = false;
                    break;
                }
                num ++;
                count = 0;
            }
        } else {
            count ++;
        }
    }

    // last data
    if ( flag_correct ) {
        if ( count > 0 && ( num != numbers.length - 1 ) ) {
            flag_correct = false;
        } else if ( count > 0 && count != numbers[ num ] ) {
            flag_correct = false;
        } else if ( count == 0 && num < numbers.length ) {
            flag_correct = false;
        }
    }
    return flag_correct;//true;
};

BoardData.prototype.getAuthor = function() {
    return this.author;
};

BoardData.prototype.setAuthor = function( author ) {
    this.author = author;
};

BoardData.prototype.getTitle = function() {
    return this.title;
};

BoardData.prototype.setTitle = function( title ) {
    this.title = title;
};

BoardData.prototype.copyPicture = function() {
    this.dots = this.answer.substring( 0, this.answer.length );
    return this.dots;
};

BoardData.prototype.getCurrentPattern = function() {
    return this.dots.replace( /\-/g, " " );
};

//BoardData.prototype.reg
// set answer pattern with size
BoardData.prototype.setPicture = function( width, height, data ) {
    var reg_hex = new RegExp( "^[0-9A-Fa-f]+$" );
    if ( reg_hex.exec( data ) ) {
        data = __decodeHexPattern( data );
    }
    // validate
    if ( width != undefined && height != undefined ) {
        this.width = width;
        this.height = height;
    }
    if ( data.length < this.width * this.height ) {
        debug( "not match" );
        return;
    }

    this.answer = data;

    // horizontal numbers
    var hnums = new Array();
    var pos = 0;
    for ( var i = 0; i < this.height; i++ ) {
        var count = 0;
        var row = new Array();
        for ( var j = 0; j < this.width; j++, pos++ ) {
            if ( data[ pos ] == " " ) {
                if ( count > 0 ) {
                    row.push( count );
                    count = 0;
                }
            } else {
                count ++;
            }
        }
        if ( count > 0 ) row.push( count );
        hnums.push( row );
    }
    this.horizontalNumbers = hnums;

    // vertical numbers
    var vnums = new Array();
    for ( var i = 0; i < this.width; i++ ) {
        var count = 0;
        var column = new Array();
        pos = i;
        for ( var j = 0; j < this.height; j ++, pos += this.width ) {
            if ( data[ pos ] == " " ) {
                if ( count > 0 ) {
                    column.push( count );
                    count = 0;
                }
            } else {
                count ++;
            }
        }
        if ( count > 0 ) column.push( count );
        vnums.push( column );
    }
    //print( vnums );
    this.verticalNumbers = vnums;

    // clear user dots
    this.dots = "";
    for ( var i = 0; i < this.answer.length; i++ ) {
        this.dots += " ";
    }
};

BoardData.prototype.getHorizontalHints = function( row ) {
    if ( row < 0 || row >= this.horizontalNumbers.length ) {
        return [];
    }
    return this.horizontalNumbers[ row ];
};

BoardData.prototype.getVerticalHints = function( col ) {
    if ( col < 0 || col >= this.verticalNumbers.length ) {
        return [];
    }
    return this.verticalNumbers[ col ];
};

// width of the board
BoardData.prototype.getWidth = function() {
    return this.width;
};

// height of the height
BoardData.prototype.getHeight = function() {
    return this.height;
};

BoardData.prototype.__getPosition = function( x, y ) {
    var pos = x;
    if ( y != undefined ) {
        pos += y * this.width;
    } else if ( x < 0 || x >= this.width ) {
        return -1;
    }

    var len = this.width * this.height;
    if ( pos < 0 || pos >= len ) {
        return - 1;
    }
    return pos;
}

BoardData.prototype.paintPanel = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;
    if ( this.dots[ pos ] != "*" ) {
        this.dots = this.dots.substring( 0, pos ) + "*" + this.dots.substring( pos + 1, this.dots.legnth );
    }
};
BoardData.prototype.clearPanel = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;
    if ( this.dots[ pos ] == "*" ) {
        this.dots = this.dots.substring( 0, pos ) + " " + this.dots.substring( pos + 1, this.dots.length );
    }
};
// toggle pixel
BoardData.prototype.togglePanel = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;

    if ( this.dots[ pos ] == " " ) {
        this.dots = this.dots.substring( 0, pos ) + "*" + this.dots.substring( pos + 1, this.dots.length );
    } else {
        this.dots = this.dots.substring( 0, pos ) + " " + this.dots.substring( pos + 1, this.dots.length );
    }
};

// whether a pixel is checked
BoardData.prototype.isChecked = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;
    return ( this.dots[ pos ] == "*" );
};

BoardData.prototype.markPanel = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;
    this.dots = this.dots.substring( 0, pos ) + "-" + this.dots.substring( pos + 1, this.dots.length );
}

// whether a pixel is marked
BoardData.prototype.isMarked = function( x, y ) {
    var pos = this.__getPosition( x, y );
    if ( pos < 0 ) return;
    return ( this.dots[ pos ] == "-" );
};


// whether a pixel is
BoardData.prototype.isFilled = function( x, y ) {
    if ( y == undefined ) y = 0;
    var pos = x + y * this.width;
    //if ( pos < 0 || pos >= this.width * this.height ) return false;
    return ( this.answer[ pos ] != " " );
}

// whether a pixel is
BoardData.prototype.isOpen = function( x, y ) {
    if ( y == undefined ) y = 0;
    var pos = x + y * this.width;
    //if ( pos < 0 || pos >= this.width * this.height ) return false;
    //debug( x + ", " + y + " = '" + this.answer[ pos ] + "'" );
    return ( this.answer[ pos ] == " " );
}

BoardData.prototype.deserialize = function( data ) {
    var start = data.length;
    var end = 0;
    for ( var i = 0; i < data.length; i++ ) {
        if ( data[ i ] == "{" ) {
            start = i;
            break;
        }
    }
    for ( var i = start; i < data.length; i++ ) {
        if ( data[ i ] == "}" ) {
            end = i;
            break;
        }
    }
    if ( end - start > 0 ) {
        var content = data.substring( start, end + 1 );
        var __temp = null;
        var evalcom = "__temp = " + content.replace( /;/g, '' );
        debug( evalcom );
        try {
            eval( evalcom );
        } catch ( e ) {
            return false;
        }
        if ( __temp == null ) return false;
        if ( __temp.width != undefined && __temp.height != undefined && __temp.pattern != undefined ) {
            this.setPicture( __temp.width, __temp.height, __temp.pattern );
            this.title = __temp.title;
            this.author = __temp.author;
            if ( this.isValid() ) {
//                 debug( "OK" );
//                 debug( "serialization : " + this.serialize() );
//                 debug( this.toString() );
                return true;
            }
        }
    }
    return false;
};

BoardData.prototype.getDensity = function() {
    var count_dots = 0;
    for ( var i = 0; i < this.dots.length; i++ ) {
        if ( this.dots[ i ] == "*" ) {
            count_dots ++;
        }
    }
    return parseFloat( count_dots ) / this.dots.length;
};

BoardData.prototype.serialize = function() {
    var text = "{width:" + this.width + ",height:" + this.height;
    if ( this.author ) {
        var author = this.author.replace( /[\"\\]/g, "" );
        text += ",author:\"" + author + "\"";
    }
    if ( this.title ) {
        var title = this.title.replace( /[\"\\]/g, "" );
        text += ",title:\"" + title + "\"";
    }
    text += ",pattern:\"" + __encodeHexPattern( this.dots );
    text += "\"}";
    return text;
};

BoardData.prototype.toString = function() {
    var output = "";
    var pos = 0;
    for ( var i = 0; i < this.height; i++ ) {
        for ( var j = 0; j < this.width; j++, pos++ ) {
            //print( pos );
            //print( this.answer );
            if ( this.answer[ pos ] != " " ) {
                output = output + "**";
            } else {
                output = output + "  ";
            }
        }
        output = output + " | ";
        var nums = this.horizontalNumbers[ i ];
        if ( nums != null && nums.length > 0 ) {
            for ( var j = 0; j < nums.length; j++ ) {
                output = output + _format( nums[ j ], 2 ) + " ";
            }
        }
        output = output + "\n";
    }

    var count = 0;
    for ( ;; count ++ ) {
        var flag_output = false;
        for ( var i = 0; i < this.width; i++ ) {
            if ( this.verticalNumbers[ i ].length > count ) {
                flag_output = true;
                output += _format( this.verticalNumbers[ i ][ count ], 2 );// + " ";
            } else {
                output += "  ";
            }
        }
        output += "\n";
        if ( flag_output == false ) break;
    }
    return output;

};


function _format( number, figure ) {
    var number = "" + number;
    if ( number.length < figure ) {
        for ( var i = number.length; i < figure; i++ ) {
            number = " " + number;
        }
    }
    return number;
}
function __decodeHexPattern( pattern ) {
    var pat = "";
    var hex = { '0':0, '1':1, '2':2, '3':3, '4':4, '5':5, '6':6, '7':7,
                '8':8, '9':9, 'A':10,'B':11,'C':12,'D':13,'E':14,'F':15 };
    pattern = pattern.toUpperCase();
    var len = pattern.length;
    try {
        for ( var i = 0; i < len; i++ ) {
            var c = pattern[ i ];
            var n = hex[ c ];
            for ( var j = 0; j < 4; j++, n <<= 1 ) {
                if ( ( n & 8 ) != 0 ) {
                    pat = "*" + pat;
                } else {
                    pat = " " + pat;
                }
            }
        }
    } catch ( e ) {
        return null;
    }
    //debug( pat );
    return pat + "   ";
}

function __encodeHexPattern( pattern ) {
    var pat = "";
    var hex = "0123456789ABCDEF";
    try {
        for ( var i = 0; i < pattern.length; i+= 4 ) {
            var n = 0;
            var mask = 1;
            for ( var j = 0; j < 4; j++, mask <<= 1 ) {
                if ( i + j >= pattern.length ) break;
                if ( pattern[ i + j ] == '*' ) {
                    n += mask;
                }
            }
            pat = hex[ n ] + pat;
        }
    } catch ( e ) {
        return null;
    }
    return pat;
}
