// class BoardLayout
// board containing elements and data
//
function _getBounds( element ) {
    //debug( "ELEM " + element );
    var left   = _getProperty( element, "left", 0 );
    var top    = _getProperty( element, "top", 0 );
    var width  =  _getProperty( element, "width", 0 );
    var height =  _getProperty( element, "height", 0 );
    return { left:left, top:top, width:width, height:height, 
            right:left + width, bottom:top + height };
}

var __regex_number = new RegExp( /^\d+/ );

function _getProperty( element, property, defaultValue ) {
    //debug( element + " " + property + " " + defaultValue );
    try {
        var style = document.defaultView.getComputedStyle( element, null );
        var value = style.getPropertyValue( property );
        var match = __regex_number.exec( value );
        if ( match ) {
            return parseInt( match[ 0 ] );
        }
    } catch ( e ) {
        debug( "cannot get property " + e );
        throw e;
    }
    return defaultValue;
}

function BoardLayout() {
    this.container                = null;
    this.canvas                   = null;
    this.horizontalHint           = null;
    this.horizontalNumberElements = null;
    this.verticalHint             = null;
    this.verticalNumberElements   = null;
    this.controlBox               = null;
    this.constructionBox          = null;
    this.dropBox                  = null;
    this.data                     = null;
    this.fontSize = 10;
    this.xstep = 1;
    this.ystep = 1;
    this.constructionMode = false;

    // appearence
    this.strokeStyle = "rgb(64,96,128)";
    this.gridStyle = "black";
    this.fillStyle = "black";
    this.markStyle = "rgb(160,192,200)";

//     this.incorrectStyle = "color:white;background:black;"
//     this.correctStyle   = "color:white;background:black;";

    this.incorrectBackgroundColor = "none";
    this.correctBackgroundColor = "darkgray";

    this.incorrectColor = "black";
    this.correctColor = "#412";

    this.canvasRatio = 0.75;
}

BoardLayout.prototype.setConstructionMode = function( flag ) {
    if ( this.constructionMode != flag ) {
        this.constructionMode = flag;
        this.refresh();
        //this.repaint();
    }
};

// refresh layout with data
BoardLayout.prototype.refresh = function() {
    if ( this.container == null || this.canvas == null || this.data == null ) {
        return;
    }

    var container = this.container;
    var boundary = _getBounds( container ); 
    var canvas = this.canvas;
    var cxt = canvas.getContext( "2d" );
    if ( cxt == null ) return null;

    // get the bounding rectangle of canvas
    var bounds = _getBounds( canvas );
    bounds.width = parseInt( boundary.width * this.canvasRatio );
    bounds.height = parseInt( boundary.height * this.canvasRatio );
    bounds.bottom = bounds.top + bounds.height;
    bounds.right  = bounds.left + bounds.width;
    canvas.style.width = bounds.width + "px";
    canvas.style.height = bounds.height + "px";

    canvas.width  = bounds.width;
    canvas.height = bounds.height;

    // division unit
    var xstep = parseInt( ( bounds.width - 1 ) / this.data.getWidth() );
    var ystep = parseInt( ( bounds.height - 1 ) / this.data.getHeight() );
    xstep = ystep = Math.min( xstep, ystep );
    this.xstep = xstep;
    this.ystep = ystep;

    // set board boundary
    var width  = this.data.getWidth();
    var height = this.data.getHeight();
    // font size
    this.fontSize = parseInt( Math.min( this.xstep, this.ystep ) * .75 );

    if ( this.constructionMode == true && this.constructionBox != null ) {
        this.constructionBox.style.display = "block";
        this.horizontalHint.style.display = "none";
        this.verticalHint.style.display = "none";
        if ( this.controlBox ) this.controlBox.style.display = "none";
        if ( this.dropBox ) this.dropBox.style.display = "block";

        this.constructionBox.style.top    = ( bounds.top + this.ystep * this.data.getHeight() ) + "px";//bounds.height + "px";
        this.constructionBox.style.left   = bounds.left + "px";
        this.constructionBox.style.width  = bounds.width + "px";
        this.constructionBox.style.height = ( boundary.height - bounds.bottom ) + "px";

        this.dropBox.style.left = bounds.right + "px";
        this.dropBox.style.top  = bounds.top + "px";
        this.dropBox.style.width = ( boundary.width - bounds.width ) + "px";
        this.dropBox.style.height = bounds.height + "px";
    } else {
        //if ( this.constructionBox ) {
        //this.constructionBox.style.display = "none";
        this.horizontalHint.style.display = "block";
        this.verticalHint.style.display = "block";
        this.controlBox.style.display = "block";
        if ( this.constructionBox ) this.constructionBox.style.display = "none";
        if ( this.dropBox ) this.dropBox.style.display = "none";

        while ( this.horizontalHint.firstChild ) {
            this.horizontalHint.removeChild( this.horizontalHint.firstChild );
        }
        while ( this.verticalHint.firstChild ) {
            this.verticalHint.removeChild( this.verticalHint.firstChild );
        }

        this.horizontalNumberElements = new Array();
        this.horizontalHint.style.top  = bounds.top + "px";
        this.horizontalHint.style.left = ( bounds.left + this.xstep * this.data.getWidth() ) + "px";
        this.horizontalHint.style.height = bounds.height + "px";
        this.horizontalHint.style.width = ( boundary.width - bounds.right ) + "px";
        
        for ( var i = 0; i < height; i++ ) {
            var div = document.createElement( "div" );
            var style = "height:" + ( this.ystep - 1 ) + "px;font-size:" + this.fontSize + "px;text-align:left;";
            div.setAttribute( "class", "horizontalNumbers" );
            div.setAttribute( "style", style );
            this.horizontalHint.appendChild( div );
            this.horizontalNumberElements.push( div );
        }
        
        var dummyrow = document.createElement( "div" );
        dummyrow.setAttribute( "class", "horizontalNumbers" );
        dummyrow.setAttribute( "style", "height:1px" );
        this.horizontalHint.appendChild( dummyrow );
        
        this.verticalNumberElements = new Array();
        this.verticalHint.style.left = bounds.left + "px";
        this.verticalHint.style.top = ( bounds.top + this.ystep * this.data.getHeight() ) + "px";
        this.verticalHint.style.width = bounds.width + "px";
        this.verticalHint.style.height = ( boundary.height - bounds.bottom ) + "px";
        for ( var i = 0; i < width; i++ ) {
            var div = document.createElement( "div" );
            var style = "width:" + ( this.xstep - 1 ) + "px;font-size:" + this.fontSize + "px;text-align:right;line-height:" + this.fontSize + "px;";
            div.setAttribute( "class", "verticalNumbers" );
            div.setAttribute( "style", style );//"width:" + ( this.xstep - 1 ) + "px;font-size:" + this.fontSize + "px;" );
            this.verticalHint.appendChild( div );
            this.verticalNumberElements.push( div );
        }
        
        
        var dummycol = document.createElement( "div" );
        dummycol.setAttribute( "class", "verticalNumbers" );
        dummycol.setAttribute( "style", "width:1px" );
        this.verticalHint.appendChild( dummycol );
    }
    //cxt.clearRect( 0, 0, canvas.width, canvas.height );

//     if ( this.constructionBox ) {
//         this.constructionBox.style.left = this.
//     }

    if ( this.controlBox ) {
        this.controlBox.style.left = ( bounds.left + this.xstep * this.data.getWidth() ) + "px";//this.horizontalHint.style.left;
        this.controlBox.style.top  = ( bounds.top + this.ystep * this.data.getHeight() ) + "px";//this.verticalHint.style.top;
        this.controlBox.style.width = ( boundary.width - bounds.right ) + "px";;//this.horizontalHint.style.width;
        this.controlBox.style.height = ( boundary.height - bounds.bottom ) + "px";//this.verticalHint.style.height;
    }
};

// set data
BoardLayout.prototype.setData = function( boardData ) {
    this.data = boardData;
    var div = document.getElementById( SizeDisplayID );
    if ( div ) {
        div.innerText = boardData.getWidth() + " x " + boardData.getHeight();
    }
    this.refresh();
};

// set HTML elements
BoardLayout.prototype.setElements = function( container, canvas, hHint, vHint, controlBox, constructionBox, dropBox ) {
    this.container = container;
    this.canvas = canvas;
    this.horizontalHint = hHint;
    this.verticalHint = vHint;
    this.controlBox = controlBox;
    this.constructionBox = constructionBox;
    this.dropBox = dropBox;
    this.refresh();
};

// repaint board and numbers
BoardLayout.prototype.repaint = function() {
    this.repaintBoard();
    this.repaintNumbers();
};

// board
BoardLayout.prototype.repaintBoard = function() {
    var canvas = this.canvas;
    if ( canvas == null || this.data == null ) return;
    var cxt = canvas.getContext( "2d" );
    if ( cxt == null ) return null;
    var bounds = _getBounds( canvas );
    cxt.clearRect( 0, 0, canvas.width, canvas.height );

    cxt.save();
    cxt.lineWidth = "1";

    var x = 0;
    var y = 0;
    var xstep = this.xstep;
    var ystep = this.ystep;
    var pos_right  = xstep * this.data.getWidth();
    var pos_bottom = ystep * this.data.getHeight();
    //cxt.strokeStyle = this.strokeStyle;

    x = 0.5; // lines are drawn at the center of the position
    for ( var i = 0; i <= this.data.getWidth(); i++ ) {
        if ( i % 5 == 0 ) {
            cxt.strokeStyle = this.gridStyle;
        } else {
            cxt.strokeStyle = this.strokeStyle;
        }
        cxt.moveTo( x, 0 );
        cxt.lineTo( x, pos_bottom );
        cxt.stroke();
        x += xstep;
    }
    y = 0.5;
    for ( var i = 0; i <= this.data.getHeight(); i++ ) {
        if ( i % 5 == 0 ) {
            cxt.strokeStyle = this.gridStyle;
        } else {
            cxt.strokeStyle = this.strokeStyle;
        }
        cxt.moveTo( 0, y );
        cxt.lineTo( pos_right, y );
        cxt.stroke();
        y += ystep;
    }

    var pos = 0;
    y = 1;
    for ( var i = 0; i < this.data.getHeight(); i++, y += ystep ) {
        x = 1;
        for ( var j = 0; j < this.data.getWidth(); j++, pos++, x += xstep ) {
            if ( this.data.dots[ pos ] == "*" ) {
                cxt.fillStyle = this.fillStyle;
                cxt.rect( x, y, xstep - 1, ystep - 1 );
                cxt.fill();
            } else if ( this.data.dots[ pos ] == "-" ) {
                cxt.fillStyle = this.markStyle;
                cxt.rect( x, y, xstep - 1, ystep - 1 );
                cxt.fill();
            }
        }
    }
    cxt.restore();
};

// set hint numbers
BoardLayout.prototype.repaintNumbers = function() {
    if ( this.constructionMode ) return;
    var data = this.data;
    var helems = this.verticalNumberElements;
    var velems = this.horizontalNumberElements;
    var rows = data.getHeight();
    var cols = data.getWidth();

    // set numbers at the right of the board
    for ( var i = 0; i < rows; i++ ) {
         var row = velems[ i ];
         while ( row.firstChild ) row.removeChild( row.firstChild );
        var nums = data.getHorizontalHints( i );
        for ( var j = 0; j < nums.length; j ++ ) {
            _addColorText( row, _format( nums[ j ], 2 ) );
        }
        this.refreshRow( i );
    }

    // set numbers at the bottom of the board
    for ( var i = 0; i < cols; i++ ) {
         var col = helems[ i ];
         while ( col.firstChild ) {
             col.removeChild( col.firstChild );
         }
        var nums = data.getVerticalHints( i );
        for ( var j = 0; j < nums.length; j++ ) {
            _addColorText( col, _format( nums[ j ], 2 ), null );
            col.appendChild( document.createElement( "br" ) );
        }
         this.refreshColumn( i );
    }
};

BoardLayout.prototype._fillWithOption = function( x1, y1, x2, y2, mode ) {//flag_mark ) {
    if ( this.data == null || this.canvas == null ) return;
    //debug( ">> " + x1 + ", " + y1 + ", " + x2 + ", " + y2 );
    if ( ( typeof x1 ) != "number" || ( typeof y1 ) != "number" 
         || ( typeof x2 ) != "number" || ( typeof y2 ) != "number" ) {
        return;
    }
    var x = parseInt( x1 / this.xstep );
    var y = parseInt( y1 / this.ystep );
    var s = parseInt( x2 / this.xstep );
    var t = parseInt( y2 / this.ystep );
    if ( mode == "Toggle" ) {
        if ( this.data.isChecked( s, t ) == false ) {
            mode = "Paint";
        } else {
            mode = "Clear";
        }
    }
    //debug( ">> " + x + ", " + y + ", " + s + ", " + t );
    if ( x < 0 || x >= this.data.getWidth() 
         || y < 0 || y >= this.data.getHeight() 
         || s < 0 || s >= this.data.getWidth() 
         || t < 0 || t >= this.data.getHeight() ) {
        return;
    }
    if ( s > x ) {
        var _x = x;
        x = s; s = _x;
    }
    if ( t > y ) {
        var _x = y;
        y = t; t = _x;
    }
    var flag_mark = ( mode == "Mark" );
    var flag_paint = ( mode == "Paint" );
    for ( var i = t; i <= y; i++ ) {
        for ( var j = s; j <= x; j++ ) {
            if ( flag_mark == true ) {
                if ( this.data.isChecked( j, i ) == false ) {
                    this.data.markPanel( j, i );
                }
            } else if ( flag_paint ) {
                if ( this.data.isMarked( j, i ) == false ) {
                    this.data.paintPanel( j, i );
                }
            } else {
                if ( this.data.isChecked( j, i ) ) {
                    this.data.clearPanel( j, i );
                }
            }
        }
    }

    this.repaintBoard();
    this.repaintNumbers();
};

BoardLayout.prototype.fillByMark = function( x1, y1, x2, y2 ) {
    this._fillWithOption( x1, y1, x2, y2, "Mark" );//true );
};

BoardLayout.prototype.fillPanel = function( x1, y1, x2, y2 ) {
    this._fillWithOption( x1, y1, x2, y2, "Toggle" );
};

BoardLayout.prototype.togglePanel = function( x0, y0 ) {
//    this.fillWithOption( x1, y1, x2, y2, false );
    if ( this.data == null || this.canvas == null ) return;
    var x = parseInt( x0 / this.xstep );
    var y = parseInt( y0 / this.ystep );
    if ( x < 0 || x >= this.data.getWidth() || y < 0 || y >= this.data.getHeight() ) {
        return;
    }
    //debug( "(x,y) = " + x + ", " + y );
    var cx = x * this.xstep;
    var cy = y * this.ystep;
    var canvas = this.canvas;
    var cxt = canvas.getContext( "2d" );
    cx += 1;
    cy += 1;
    cxt.clearRect( cx, cy, this.xstep - 1, this.ystep - 1 );
    //debug( this.data.dots );//isFilled( x, y ) );//Checked( x, y ) );
    this.data.togglePanel( x, y );
    //debug( this.data.dots );//isFilled( x, y ) );//Checked( x, y ) );
    if ( this.data.isChecked( x, y ) ) {
        cxt.fillStyle = this.fillStyle;
        cxt.rect( cx, cy, this.xstep - 1, this.ystep - 1 );
        cxt.fill();
    } 

    this.refreshRow( y );
    this.refreshColumn( x );
};


BoardLayout.prototype.markPanel = function( x, y ) {
    if ( this.data == null || this.canvas == null ) return;
    var x = parseInt( x / this.xstep );
    var y = parseInt( y / this.ystep );
    if ( x < 0 || x >= this.data.getWidth() || y < 0 || y >= this.data.getHeight() ) {
        return;
    }
    //debug( x + ", " + y );
    var cx = x * this.xstep;
    var cy = y * this.ystep;
    var canvas = this.canvas;
    var cxt = canvas.getContext( "2d" );
    cx += 1;
    cy += 1;
    cxt.clearRect( cx, cy, this.xstep - 1, this.ystep - 1 );
    if ( this.data.isChecked( x, y ) ) {
        this.data.togglePanel( x, y );
    } 
    this.data.markPanel( x, y );
    cxt.fillStyle = this.markStyle;//"lightGray";//this.fillStyle;
    cxt.rect( cx, cy, this.xstep - 1, this.ystep - 1 );
    cxt.fill();

    this.refreshRow( y );
    this.refreshColumn( x );
};

function _addColorText( parent, text, style ) {
    var span = document.createElement( "span" );
    if ( style != undefined ) {
        span.setAttribute( "style", style );
    }
    span.appendChild( document.createTextNode( text ) );
    parent.appendChild( span );
}

BoardLayout.prototype.refreshRow = function( row ) {
    if ( this.constructionMode ) return;
    var flag_correct = this.data.rowMatches( row );
    var elem    = this.horizontalNumberElements[ row ];
    if ( flag_correct ) {
        elem.style.backgroundColor = this.correctBackgroundColor;
        elem.style.color = this.correctColor;
    } else {
        elem.style.backgroundColor = this.incorrectBackgroundColor;
        elem.style.color = this.incorrectColor;
    }
    return flag_correct;
};

BoardLayout.prototype.refreshColumn = function( col ) {
    if ( this.constructionMode ) return;
    var elem    = this.verticalNumberElements[ col ];
    var flag_correct = this.data.columnMatches( col );
    
    if ( flag_correct ) {
        elem.style.backgroundColor = this.correctBackgroundColor;
        elem.style.color = this.correctColor;
    } else {
        elem.style.backgroundColor = this.incorrectBackgroundColor;
        elem.style.color = this.incorrectColor;
    }

    return flag_correct;
};

