// access to HTML/CSS
var ContainerID         = "boardContainer";
var BoardElementID      = "boardCanvas";
var HorizontalNumbersID = "horizontalHint";
var VerticalNumbersID   = "verticalHint";
var ControlBoxID        = "controlBox";
var MessagePanelID      = "messagePanel";
var ConstructionBoxID   = "constructionBox";
var DropBoxID           = "dropBox";
var StatusBarID         = "statusBar";
var waitRollerID        = "wait";
var QuizNumberID        = "quizNumber";
var SizeDisplayID       = "currentSize";
var startTime = 0;
var lockFlag = false;
var autoloadFlag = false;

// data
var boardData = null;
var boardLayout = new BoardLayout();
var QuizNumber = 0;
var QuizSet = Quiz;//_16;
var commands = null;
var userName = null;

var playingMode = "Challenge"; // "Challenge or Construction"

// parameter containing elements
var authorPrefID = "authorName";
var titlePrefID = "pictureTitle";
var autoloadPrefID = "autoload";
var modePrefID_challenge = "challengeMode";
var modePrefID_construction = "constructionMode";

var HorizontalNumberClass = "horizontalNumbers";
var VerticalNumberClass   = "verticalNumbers";

var logging = true;
function debug( message ) {
    if ( logging == false ) return;
    if ( message == null ) message = "Null";
    if ( message == undefined ) message = "Undefined";

    if ( window.SillyImageConverter ) {
        SillyImageConverter.log( message.toString() );
    } else {
        var div = document.getElementById( "debug" );
        if ( div ) {
            div.appendChild( document.createTextNode( message ) );
            div.appendChild( document.createElement( "br" ) );
        }
    }
}


function tr( key, defvalue ) {
    if ( defvalue == null ) defvalue = key;
    try {
        var localized = localizedStrings[ key ];
        if ( localized ) {
            return localized;
        }
        return defvalue;
    } catch ( e ) {
        return defvalue;
    }
}

// setup
function init() {
    // preferences
    __importPreferences();

    // localization
    var elements = document.getElementsByTagName( "SPAN" );
    for ( var i = 0; i < elements.length; i++ ) {
        var elem = elements[ i ];
        if ( elem.localizationKey ) {
            elem.innerText = tr( elem.localizationKey );
        }
    }

    // read stored quizes from sqlite database
    if ( window.SillyImageConverter ) {
        var homeDir = SillyImageConverter.getHomeDirectory();
        debug( homeDir );
        var db_dir = homeDir + "/Library/Application Support/Pixelotium/";
        var db_filename = db_dir + "pixelotium.db";
        debug( db_dir );
        debug( db_filename );
        // create application support directory
        widget.system( "/bin/mkdir -p \"" + db_dir + "\"", null );

        SillyImageConverter.setDBFilename( db_filename );

        var pats = SillyImageConverter.readPictures();
        for ( var i = 0; i < pats.length; i++ ) {
            var q = new Array();
            q.width = parseInt( pats[ i ].width );
            q.height = parseInt( pats[ i ].height );
            q.pattern = pats[ i ].pattern;
            if ( pats[ i ].author.length > 0 ) {
                q.author = pats[ i ].author;
            }
            if ( pats[ i ].title.length > 0 ) {
                q.title = pats[ i ].title;
            }
            var id = pats[ i ].id;
            QuizID[ id ] = q;
            Quiz.push( q );
        }
    }

    boardData = new BoardData();

    boardLayout = new BoardLayout();
    boardLayout.setElements( document.getElementById( ContainerID),
                             document.getElementById( BoardElementID ),
                             document.getElementById( HorizontalNumbersID ), 
                             document.getElementById( VerticalNumbersID ),
                             document.getElementById( ControlBoxID ),
                             document.getElementById( ConstructionBoxID ),
                             document.getElementById( DropBoxID )
        );
    boardLayout.setData( boardData );
    _showQuiz( 0 );
    //boardLayout.repaint();
    commands = new CommandManager( boardData, boardLayout );

    createResizer( document.getElementById( ContainerID ), 5, -10, function() { boardLayout.refresh(); boardLayout.repaint(); }, {width:150,height:150},{width:700,height:700} );

    // functional buttons
    __createButtons();

    // 'i' button
    createInformationButton( document.getElementById( "front" ), 12, -22, flipToBehind, true );

    document.addEventListener( "keypress", __keypress, true );
    

    // loading timer images
    __prepareWaitTimer();

    // load from network
    if ( autoloadFlag ) {
        setTimeout( loadFromBlog, 250 );
    }

}


function __createButtons() {
    // 'done' button on reverse
    createGenericButton( document.getElementById( "done" ), tr( "Done" ), flipToFront );

    // construction tools
    createButton( document.getElementById( "moveNext" ), ">", new ButtonDelegation( moveForward ) );
    createButton( document.getElementById( "movePrevious" ), "<", new ButtonDelegation( moveBackward ) );
    createButton( document.getElementById( "postButton" ), tr( "POST" ), new ButtonDelegation( postToBlog ) );
    createButton( document.getElementById( "loadButton" ), tr( "LOAD" ), new ButtonDelegation( loadFromBlog ) );
    createButton( document.getElementById( "saveButton" ), tr( "SAVE" ), new ButtonDelegation( saveToDB ) );

    createButton( document.getElementById( "expandH" ), ">", new ResizeButtonListener( 1, 0 ) );
    createButton( document.getElementById( "shrinkH" ), "<", new ResizeButtonListener( -1, 0 ) );
    createButton( document.getElementById( "shrinkV" ), "^", new ResizeButtonListener( 0, -1 ) );
    createButton( document.getElementById( "expandV" ), "v", new ResizeButtonListener( 0, 1 ) );

}

function __getPreference( key, defval ) {
    if ( defval == undefined ) defval = key;
    try {
        var val = widget.preferenceForKey( key );
        if ( val ) return val;
        return defval;
    } catch ( e ) {
    }
    return defval;
}
function __setPreference( key, value ) {
    if ( window.widget ) {
        widget.setPreferenceForKey( value, key );
    }
}

var PrefKey_Author = "AUTHOR";
var PrefKey_Title  = "TITLE";
var PrefKey_Autoload = "AUTOLOAD";

function changeAuthor( event, elem ) {
    //debug( "change author " + elem.value );
    if ( elem.value ) __setPreference( PrefKey_Author, elem.value );
}

function __importPreferences() {
    {
        var div = document.getElementById( authorPrefID );
        if ( div ) div.value = __getPreference( PrefKey_Author, "" );
    }
    {
        var div = document.getElementById( titlePrefID );
        if ( div ) div.value = __getPreference( PrefKey_Title, "" );
    }
    {
        var div = document.getElementById( autoloadPrefID );
        var flag = __getPreference( PrefKey_Autoload, autoloadFlag );
        if ( flag == true ) {
            if ( div ) div.setAttribute( "checked", true );
            autoloadFlag = true;
        } else {
            if ( div ) div.removeAttribute( "checked" );
            autoloadFlag = false;
        }
    }
}

function __getCurrentAuthor() {
    var div = document.getElementById( authorPrefID );
    if ( div ) {
        return div.value;
    } else {
        return "";
    }
}
function __getCurrentTitle() {
    var div = document.getElementById( titlePrefID );
    if ( div ) {
        return div.value;
    } else {
        return "";
    }
}

function __remove() {
}



function CommandManager( data, layout ) {
    this.boardData   = data;
    this.boardLayout = layout;
    this.pos = 0;
    this.maxSize = 32;
    this.history = new Array();
    this.toggleCommand = "Toggle";
    this.fillCommand = "Fill";
    this.markCommand = "Mark";
    this.lastPosition = null;
    
    for ( var i = 0; i < this.maxSize; i++ ) {
        this.history[ i ] = null;
    }

    this.toggle = function( x, y ) {
        this.push();
        this.boardLayout.togglePanel( x, y );
        this.lastPosition = { x:x, y:y };
    };
    this.mark = function( x, y ) {
        this.push();
        this.boardLayout.markPanel( x, y );
        this.lastPosition = { x:x, y:y };
    };
    this.fillPanel = function( x, y ) {
        if ( this.lastPosition ) {
            this.push();
            this.boardLayout.fillPanel( this.lastPosition.x, this.lastPosition.y, x, y );
            this.lastPosition = { x:x, y:y };
        }
    };
    this.fillByMark = function( x, y ) {
        if ( this.lastPosition ) {
            this.push();
            this.boardLayout.fillByMark( this.lastPosition.x, this.lastPosition.y, x, y );
            this.lastPosition = { x:x, y:y };
        }
    };
    this.push = function( keepFlag ) {
        if ( this.pos >= this.maxSize ) {
            for ( var i = 0; i < this.history.length - 1; i++ ) {
                this.history[ i ] = this.history[ i + 1 ];
            }
            this.pos --;
        }

        if ( this.history.length <= this.pos ) {
            this.history.push( this.boardData.dots.substring( 0, this.boardData.dots.length ) );
        } else {
            this.history[ this.pos ] = this.boardData.dots.substring( 0, this.boardData.dots.length );
        }

        this.pos ++;
        if ( keepFlag == false || keepFlag == undefined ) {
            for ( var i = this.pos; i < this.maxSize; i++ ) {
                if ( this.history[ i ] == null ) break;
                this.history[ i ] = null;
            }
        }
    };
    this.undo = function() {
        if ( this.pos == 0 ) return;
        this.push( true );
        this.pos -= 2;
        this.boardData.dots = this.history[ this.pos ];
        this.boardLayout.refresh();
        //this.boardLayout.setData( this.boardData );
        this.boardLayout.repaint();
    };
    this.redo = function() {
        if ( this.history[ this.pos + 1 ] == null ) return;
        this.pos ++;
        this.boardData.dots = this.history[ this.pos ];
        this.boardLayout.refresh();
        //this.boardLayout.setData( this.boardData );
        this.boardLayout.repaint();
    };
    this.clearBuffer = function() {
        for ( var i = 0; i < this.maxSize; i++ ) {
            this.history[ i ] = null;
        }
        this.pos = 0;
    };
}

function __lockBoard() {
    lockFlag = true;
}

function __unlockBoard() {
    lockFlag = false;
}

function _showQuiz( num ) {
    if ( num < 0 ) num = QuizSet.length - 1;
    if ( num >= QuizSet.length ) num = 0;
    QuizNumber = num;
    hideMessage();
    __unlockBoard();
    var quiz = QuizSet[ QuizNumber ];
//    debug( "Q: " + num + " " + quiz );
//     for ( var p in quiz ) {
//         debug( p + ": " + quiz[ p ] );
//     }
    if ( quiz.width != undefined && quiz.height != undefined && quiz.pattern != undefined ) {
        var pattern = quiz.pattern;
//         var reg_pat = new RegExp( "^[0-9A-Fa-f]+$" );
//         if ( reg_pat.exec( pattern ) ) {
//             pattern = __decodeHexPattern( pattern );
//         } else {
//             debug( __encodeHexPattern( pattern ) );
//         }
        //debug( "'" + pattern + "'" );
        
        boardData.setPicture( quiz.width, quiz.height, pattern );
        boardData.setTitle( quiz.title );
        boardData.setAuthor( quiz.author );
        //debug( boardData.serialize() );

        if ( commands ) commands.clearBuffer();
        boardLayout.setData( boardData );
        boardLayout.repaint();
        startTime = (new Date()).getTime();

        var numdiv = document.getElementById( QuizNumberID );
        numdiv.innerText = ( QuizNumber + 1 ) + " / " + QuizSet.length;

        return true;
    } else {
        throw "invalid quiz " + QuizSet[ QuizNumber ];
    }
}

function moveBackward(event) {
    _showQuiz( QuizNumber - 1 );
}

function moveForward(event) {
    _showQuiz( QuizNumber + 1 );
}

function ButtonDelegation( handler ) {
    this.handler = handler;
    this.actionPerformed = function() {
        this.handler();
    }
}

function __changeMode( mode_construction ) {
    var mode = playingMode;
    if ( mode_construction == true ) {
        mode = "Construction";
    } else if ( mode_construction == false ) {
        mode = "Challenge";
    }
    debug( mode + " " + mode_construction );
    if ( mode != playingMode ) {
        boardLayout.setConstructionMode( mode_construction );
        playingMode = mode;
    }
}



function flipToBehind( event ) {
    var front = document.getElementById( "front" );
    var behind = document.getElementById( "behind" );
    if ( window.widget ) {
        widget.prepareForTransition( "ToBack" );
    }
    behind.style.display = "block";
    front.style.display = "none";
    if ( window.widget ) {
        setTimeout( "widget.performTransition();", 0 );
    }
}
function flipToFront( event ) {
    var front = document.getElementById( "front" );
    var behind = document.getElementById( "behind" );
    if ( window.widget ) {
        widget.prepareForTransition( "ToFront" );
    }

    var autoload = document.getElementById( autoloadPrefID );
    if ( autoload.value == "on" ) {
        __setPreference( PrefKey_Autoload, true );
        autoloadFlag = true;
    } else {
        __setPreference( PrefKey_Autoload, false );
        autoloadFlag = false;
    }
//}
    //debug( "Autoload " + autoload.value + ", " + autoload.getAttribute( "checked" ) + ", " + autoload.checked );

    behind.style.display = "none";
    front.style.display = "block";

    var pref_cnst = document.getElementById( "constructionMode" );
    if ( pref_cnst.checked ) {
        __changeMode( true );
        hideMessage();
    } else {
        __changeMode( false );
    }
    boardLayout.repaint();
    if ( window.widget ) {
        setTimeout( "widget.performTransition();", 0 );
    }
}

function postToBlog( event ) {
    __postPattern();
}

function showStatus( message ) {
    var div = document.getElementById( StatusBarID );
    if ( div ) {
        div.innerText = message;
    }
}

function saveToDB( event ) {
    if ( window.SillyImageConverter != undefined && window.widget != undefined ) {
        var id = "PI:" + (new Date()).getTime();
        var ttl = __getCurrentTitle();
        var ath = __getCurrentAuthor();
        var w = boardData.getWidth();
        var h = boardData.getHeight();
        var pat = __encodeHexPattern( boardData.getCurrentPattern() );
        SillyImageConverter.insertPicture( id, ttl, ath, w, h, pat );

        var q = { width:w, height:h, pattern:pat, author:ath, title:ttl};
        QuizID[ id ] = q;
        Quiz.push( q );
        
    }
//    debug( "save" );
}

function loadFromBlog( event ) {
    startWaitTimer();
    var loader = new SiteLoader( "http://blog.livedoor.jp/melmiffy/atom.xml?" + ( new Date()).getTime() , __parseRSS, __errorHandler );
    loader.xml = true;
    showStatus( tr( "Loading.." ) );
    debug( "loading" );
    startWaitTimer();
    loader.start();
}

function __errorHandler( text ) {
    debug( text );
    showStatus( tr( "Load Error" ) );
    stopWaitTimer();
}

function __parseRSS( xml, loader ) {
    debug( "loaded" );
    stopWaitTimer();
    var feed  = xml.firstChild;//
    //outputDOMText( feed, debug );
//getElementAt( document.firstChild, "feed" );
    var entry = getElementAt( feed, "entry" );
    var num = 1;
    var num_added = 0;
    var loaded = new Array();
    while ( entry ) {
        var id = getInsideTexts( entry, '', "id" );
        var pattern = getInsideTexts( entry, '', "summary" );
        var bd = new BoardData();
        if ( QuizID[ id ] == undefined ) {
            if ( bd.deserialize( pattern ) ) {
                //debug( bd.toString() );//serialize() );
                bd.id = id;
                //debug( bd.copyPicture() );
                var dict = {width:bd.getWidth(),height:bd.getHeight(),author:bd.getAuthor(),title:bd.getTitle(),pattern:bd.copyPicture()};
                loaded.push( dict );
                QuizID[ id ] = dict;
                //Quiz.push( dict );
                if ( window.SillyImageConverter ) {
                    var ttl = bd.getTitle();
                    var ath = bd.getAuthor();
                    var w = bd.getWidth();
                    var h = bd.getHeight();
                    if ( ttl == undefined ) ttl = "";
                    if ( ath == undefined ) ath = "";
                    if ( w != undefined && h != undefined ) {
                        var coded_pat = __encodeHexPattern( bd.copyPicture() );
                        //debug( coded_pat );
                        debug( "ID = " + id );
                        SillyImageConverter.insertPicture( id, ttl, ath, w, h, coded_pat );
                    }
                }
                num_added ++;
            }
        } else {
            debug( "same pattern exists" );
        }
//         debug( id );
//         debug( pattern );
        //outputDOMText( entry, debug );
        num ++;
        entry = getElementAt( feed, "entry," + num );
    }
    if ( num_added > 0 ) {
        for ( var i = loaded.length - 1; i >= 0; i-- ) {
            Quiz.push( loaded[ i ] );
        }
        showStatus( tr( "Loaded" ) + " " + num_added + " " + tr( "Pictures" ) );
    } else {
        showStatus( tr( "No New Pictures" ) );
    }
    //outputDOMText( xml.firstChild, debug );
}

function __postPattern( data ) {
    if ( data == undefined ) {
        data = boardData;
    }
    if ( data.getDensity() < 0.3 ) {
        showStatus( tr( "Too Few Dots" ) );
        return;
    }
    showStatus( tr( "Posting a Picture" ) );
    data.setAuthor( __getCurrentAuthor() );
    data.setTitle( __getCurrentTitle() );
    
    if ( window.widget ) {
        var command = widget.system( "/usr/sbin/sendmail -t", __endHandler );
        command.write( "From:melmiffy@sheep.livedoow.com\n" );
        command.write( "To:b-5da3dafd@blog.livedoor.com\n" );
        command.write( "Subject:Pixelotium\n\n" );
        debug( data.serialize() );
        command.write( data.serialize() );
        command.write( "\n\n.\n" );
        command.close();
    }
}

function __endHandler() {
    debug( "mail end" );
    showStatus( tr( "Posted" ) );
}

function __keypress( event ) {
    var code = event.keyCode;
    var meta = event.metaKey;
    var shift = event.shiftKey;
    var ctrl  = event.ctrlKey;

    // undo / redo
    if ( code == 122 && meta ) {
        if ( shift ) { // redo
            if ( commands != null ) {
                commands.redo();
            }
        } else { // undo
            if ( commands != null ) {
                commands.undo();
            }
        }
        event.preventDefault();
        event.stopPropagation();
        return;
    }

    // inverse
    if ( ctrl == true && code == 105 ) {
        for ( var i = 0; i < boardData.getHeight(); i++ ) {
            for ( var j = 0; j < boardData.getWidth(); j++ ) {
                boardData.togglePanel( j, i );
            }
        }
        boardLayout.repaint();
        event.preventDefault();
        event.stopPropagation();
        return;
    }

    // show answer
    if ( meta == true && code == 99 ) {
        boardData.copyPicture();
        boardLayout.repaint();
        
        event.preventDefault();
        event.stopPropagation();
        return;
    }

    // clear all
    if ( ctrl == true && code == 100 ) {
        if ( commands ) commands.push();
        boardData.setPicture( boardData.getWidth(), boardData.getHeight(), boardData.answer );
        boardLayout.setData( boardData );
        boardLayout.repaint();

        event.preventDefault();
        event.stopPropagation();
        return;
    }

    if ( boardData == null ) return;

    // resize panel
    var width = boardData.getWidth();
    var height = boardData.getHeight();
    if ( event.ctrlKey ) {
        if ( code == 63235 ) {
            width ++;
        } else if ( code == 63234 ) {
            width --;
        } else if ( code == 63232 ) {
            height --;
        } else if ( code == 63233 ) {
            height ++;
        } else {
            return;
        }
    } else {
        return;
    }

    if ( width <= 0 || height <= 0 ) return;
    var pattern_new = "";
    for ( var i = 0; i < width * height; i++ ) {
        if ( Math.random() < 0.5 ) {
            pattern_new += " ";
        } else {
            pattern_new += "*";
        }
    }

    boardData.setPicture( width, height, pattern_new );
    boardLayout.setData( boardData );
    boardLayout.repaint();
    event.stopPropagation();
    event.preventDefault();
}

function ResizeButtonListener( dw, dh ) {
    this.dw = dw;
    this.dh = dh;
    this.actionPerformed = function() {
        // resize panel
        var width = boardData.getWidth() + this.dw;
        var height = boardData.getHeight() + this.dh;
        
        if ( width <= 0 || height <= 0 || width > 50 || height > 50 ) return;
        var pattern_new = "";
        var picture = boardData.getCurrentPattern();

        for ( var i = 0; i < height; i++ ) {
            for ( var j = 0; j < width; j++ ) {
                if ( i < boardData.getHeight() && j < boardData.getWidth() ) {
                    pattern_new += picture[ i * boardData.getWidth() + j ];
                } else {
                    pattern_new += " ";
                }
            }
        }

        boardData.setPicture( width, height, pattern_new );
        boardData.copyPicture();
        boardLayout.setData( boardData );
        boardLayout.repaint();
    };
}

// drag
function draggingOver(event ) {
    event.stopPropagation();
    event.preventDefault();
}

var __mode_HTML = false;
var dragStartPoint = { x:0, y:0 };
function dragStart( event ) {
    var img = document.getElementById( "_hidden_dragIcon" );
    event.dataTransfer.setDragImage( img, 24, 24 );

    if ( __mode_HTML == false ) {

        var text = boardData.serialize();
        event.dataTransfer.effectAllowed = "copy";
        event.dataTransfer.setData( "Text", text );
    } else {
        event.dataTransfer.effectAllowed = "copy";
        var url = "file:///Users/tendo/Desktop/usako.png";
        event.dataTransfer.setData( "URL", url );
    }
    dragStartPoint = { x:event.screenX, y:event.screenY };

    window.addEventListener( "drag", __dragQuiz, true );
    window.addEventListener( "dragend", __dragQuizEnd, true );

}
function __dragQuiz( event ) {
    var x = event.screenX;
    var y = event.screenY;

    if ( Math.abs( x - dragStartPoint.x ) > 200 || Math.abs( y - dragStartPoint.y ) > 200 ) {
        if ( window.widget ) {
            widget.openApplication( "com.apple.dock" );
        }
    }
}

function __dragQuizEnd( event ) {
    //debug( "drag end" );
    window.removeEventListener( "drag", __dragQuiz, true );
    window.removeEventListener( "dragend", __dragQuizEnd, true );
    event.stopPropagation();
    event.preventDefault();
}

function dropImage( event, element ) {
    var uri;
    uri = event.dataTransfer.getData("text/uri-list");
    debug( "URL : " + uri );
    if ( uri.toLowerCase().lastIndexOf( ".gif" ) < 0
         && uri.toLowerCase().lastIndexOf( ".png" ) < 0
         && uri.toLowerCase().lastIndexOf( ".jpg" ) < 0
         && uri.toLowerCase().lastIndexOf( ".jpeg" ) < 0 ) {
        showStatus( tr( "Not an Image" ) );
        debug( "not image" );
        return;
    }


    if ( window.SillyImageConverter ) {
        debug( "calling plugin" );
        if ( element ) {
            element.src = uri;
        }
        var obj = SillyImageConverter.convertImage( uri, boardData.getWidth(), boardData.getHeight() );//32, 32 );
        if ( obj ) {
            debug( obj.pattern );
            boardData.setPicture( boardData.getWidth(), boardData.getHeight(), obj.pattern );
            boardData.copyPicture();
            boardLayout.repaint();
        } else {
            debug( "no available data" );
        }
    }

    event.stopPropagation();
    event.preventDefault();
}

function createResizer( resizableElement, right, bottom, resizeHandler, minsize, maxsize, keepsratio ) {
    var img = document.createElement( "img" );
    img.src = "/System/Library/WidgetResources/resize.png";
    img.width = "12";
    img.height = "12";

    if ( right == undefined ) right = 10;
    if ( bottom == undefined ) bottom = 10;

    //debug( right + ", " + bottom );
    img.setAttribute( "style", 
                      "-apple-dashboard-region:dashboard-region(control rectangle);"
                      + "width:12px;height:12px;position:absolute;"
                      + "bottom:" + bottom + "px;right:" + right + "px;"
        );
    img.setAttribute( "onmousedown", "startResizing(event,this)" );
    img.target = resizableElement;
    img.ratioFlag = keepsratio;
    if ( minsize != undefined && minsize.width != undefined && minsize.height != undefined ) {
        img.minSize = minsize;
    } else {
        img.minSize = { width:0, height:0 };
    }
    if ( maxsize != undefined && maxsize.width != undefined && maxsize.height != undefined ) {
        img.maxSize = maxsize;
    } else {
        img.maxSize = { width:0, height:0 };
    }
    if ( resizeHandler == undefined ) {
        resizeHandler = null;
    }
    img.resizeHandler = resizeHandler;

    resizableElement.appendChild( img );
}

var __movingResizer = null;
function startResizing( event, element ) {
    window.addEventListener( "mousemove", _resizeWindow, true );
    window.addEventListener( "mouseup", _stopResizing, true );
    var bounds = _getBounds( element.target );
    element.boundingRect = bounds;
    __movingResizer = element;
    __movingResizer.lastPosition = { x:event.screenX, y:event.screenY };
}

function _resizeWindow( event ) {
    if ( __movingResizer == null ) return;
    var bounds = __movingResizer.boundingRect;
    var x = event.screenX;
    var y = event.screenY;
    var dx = x - __movingResizer.lastPosition.x;
    var dy = y - __movingResizer.lastPosition.y;

    if ( __movingResizer.maxSize ) {
        if ( dx + bounds.width > __movingResizer.maxSize.width ) {
            dx = __movingResizer.maxSize.width - bounds.width;
        }
        if ( dy + bounds.height > __movingResizer.maxSize.height ) {
            dy = __movingResizer.maxSize.height - bounds.height;
        }
    }

    if ( __movingResizer.minSize ) {
        if ( dx + bounds.width < __movingResizer.minSize.width ) {
            dx = __movingResizer.minSize.width - bounds.width;
        }
        if ( dy + bounds.height < __movingResizer.minSize.height ) {
            dy = __movingResizer.minSize.height - bounds.height;
        }
    }

    __movingResizer.lastPosition = { x:x, y:y };
    window.resizeBy( dx, dy );
    var elem = __movingResizer.target;

    bounds.width  += dx;
    bounds.height += dy;

    elem.style.width = bounds.width + "px";
    elem.style.height = bounds.height + "px";

    if ( __movingResizer.resizeHandler != null ) {
        __movingResizer.resizeHandler();
    }

}

function _stopResizing( event ) {
    window.removeEventListener( "mousemove", _resizeWindow, true );
    window.removeEventListener( "mouseup", _stopResizing, true );
    __movingResizer = null;
}

// receive click event
function boardClicked( event ) {
    return;
    if ( lockFlag ) return;
    if ( boardLayout != null ) {
        if ( event.metaKey ) {
            //debug( "meta" );
            if ( event.shiftKey ) {
                commands.fillByMark( event.layerX, event.layerY );
            } else {
                commands.fillPanel( event.layerX, event.layerY );
            }
        } else {
//             if ( event.shiftKey ) {
//                 commands.mark( event.layerX, event.layerY );
//                 //boardLayout.markPanel( event.layerX, event.layerY );
//             } else {
//                 commands.toggle( event.layerX, event.layerY );
//                 //boardLayout.togglePanel( event.layerX, event.layerY );
//             }
        }

        if ( playingMode == "Challenge" && boardData.completes() ) {
            showMessage();
        } else {
            hideMessage();
        }
    }
}

function hideMessage() {
    var div = document.getElementById( MessagePanelID );
    if ( div ) {
        div.style.display = "hide";
    }
}

function _createMessageRow( header, value ) {
    var row = document.createElement( "tr" );
    {
        var td = document.createElement( "td" );
        td.setAttribute( "style", "text-align:right" );
        td.appendChild( document.createTextNode( header ) );
        row.appendChild( td );
    }
    {
        var td = document.createElement( "td" );
        td.setAttribute( "style", "text-align:left" );
        td.appendChild( document.createTextNode( value ) );
        row.appendChild( td );
    }
    return row;
}

function showMessage() {
    var div = document.getElementById( MessagePanelID );
    if ( div ) {
        __lockBoard();
        while ( div.firstChild ) div.removeChild( div.firstChild );
        var title = document.createElement( "div" );
        title.appendChild( document.createTextNode( "Congratulations!" ) );
        div.appendChild( title );
        var center = document.createElement( "center" );
        var table = document.createElement( "table" );
        var time_elapsed = parseInt( ( ( (new Date()).getTime() - startTime ) / 1000 ) );
        var min = parseInt( time_elapsed / 60 );
        var sec = time_elapsed % 60;
        table.appendChild( _createMessageRow( "Time", min + "min " + sec + "sec" ) );

        var bt = boardData.getTitle();
        debug( "title " + boardData.getTitle() );
        if ( bt != undefined && bt.length > 0 ) {
            table.appendChild( _createMessageRow( "Title", bt ) );
        }
        var ba = boardData.getAuthor();
        debug( "author " + boardData.getAuthor() );
        if ( ba != undefined && ba.length > 0 ) {
            table.appendChild( _createMessageRow( "Author", ba ) );
        }
        center.appendChild( table );
        div.appendChild( center );
        div.style.display = "block";
    }
}

var lastPaintedPanel = null;

function boardMouseMove( event ) {
    var x = parseInt( event.layerX / boardLayout.xstep ) + 1;
    var y = parseInt( event.layerY / boardLayout.ystep ) + 1;
    if ( x > 0 & x <= boardData.getWidth() && y > 0 && y <= boardData.getHeight() ) {
        showStatus( "( " + x + ", " + y + ")" );
    }
}

function boardMouseDown(event) {
    if ( lockFlag ) return;
    document.addEventListener( "mousemove", __boardMousePaint, true );
    document.addEventListener( "mouseup", __boardMouseUp, true );
    var x = parseInt( event.layerX / boardLayout.xstep );
    var y = parseInt( event.layerY / boardLayout.ystep );
        if ( event.shiftKey ) {
            commands.mark( event.layerX, event.layerY );
        } else {
            commands.toggle( event.layerX, event.layerY );
        }
    lastPaintedPanel = {x:x, y:y};
}
function __boardMousePaint( event ) {
    var x = parseInt( event.layerX / boardLayout.xstep );
    var y = parseInt( event.layerY / boardLayout.ystep );

    if ( lastPaintedPanel.x != x || lastPaintedPanel.y != y ) {
        if ( event.shiftKey ) {
            commands.mark( event.layerX, event.layerY );
        } else {
            commands.toggle( event.layerX, event.layerY );
        }

        lastPaintedPanel.x = x;
        lastPaintedPanel.y = y;
    }
}
function __boardMouseUp(event) {
    document.removeEventListener( "mousemove", __boardMousePaint, true );
    document.removeEventListener( "mouseup", __boardMouseUp, true );
    lastPaintedPanel = null;

    if ( playingMode == "Challenge" ) {
        if ( boardData.completes() ) {
            showMessage();
        } else {
            hideMessage();
        }
    }
}

var __waitTimer = { count:0, steps:12, time:66, element:null, timer:null };

function __prepareWaitTimer() {
    for ( var i = 0; i < __waitTimer.steps; i++ ) {
        ( new Image() ).src = "Images/Wait" + i + ".png";
    }
    __waitTimer.element = document.getElementById( waitRollerID );
}

function startWaitTimer() {
    if ( __waitTimer.element == null ) return;
    if ( __waitTimer.timer != null ) return;
    __waitTimer.element.style.display = "block";
    __waitTimer.timer = setInterval( __tickWaitTimer, __waitTimer.time );
    __tickWaitTimer();
}

function __tickWaitTimer() {
    __waitTimer.element.src = "Images/Wait" + __waitTimer.count + ".png";
    __waitTimer.count = ( __waitTimer.count + 1 ) % __waitTimer.steps;
}

function stopWaitTimer() {
    __waitTimer.element.style.display = "none";
    clearInterval( __waitTimer.timer );
    __waitTimer.timer = null;
}
